@@ -16,6 +16,13 @@ import copy
 
 import pandas
 
+import numpy as np
+
+import datetime
+from datetime import timedelta
+
+import pandas as pd
+
 from findatapy.market.ioengine import IOEngine
 from findatapy.market.marketdatarequest import MarketDataRequest
 from findatapy.timeseries import Filter, Calculations
@@ -32,7 +39,6 @@ class MarketDataGenerator(object):
 
     def __init__(self):
         self.config = ConfigManager().get_instance()
-        self.logger = LoggerManager().getLogger(__name__)
         self.filter = Filter()
         self.calculations = Calculations()
         self.io_engine = IOEngine()
@@ -44,102 +50,107 @@ class MarketDataGenerator(object):
     def set_intraday_code(self, code):
         self._intraday_code = code
 
-    def get_data_vendor(self, source):
+    def get_data_vendor(self, md_request):
         """Loads appropriate data service class
 
         Parameters
         ----------
-        source : str
-            the data service to use "bloomberg", "quandl", "yahoo", "google", "fred" etc.
+        md_request : MarketDataRequest
+            the data_source to use "bloomberg", "quandl", "yahoo", "google", "fred" etc.
             we can also have forms like "bloomberg-boe" separated by hyphens
 
         Returns
         -------
         DataVendor
         """
+        logger = LoggerManager().getLogger(__name__)
 
-        data_vendor = None
+        data_source = md_request.data_source
+        data_engine = md_request.data_engine
 
-        try:
-            source = source.split("-")[0]
-        except:
-            self.logger.error("Was data source specified?")
-
-            return None
-
-        if source == 'bloomberg':
-            try:
-                from findatapy.market.datavendorbbg import DataVendorBBGOpen
-                data_vendor = DataVendorBBGOpen()
-            except:
-                self.logger.warn("Bloomberg needs to be installed")
-
-        elif source == 'quandl':
-            from findatapy.market.datavendorweb import DataVendorQuandl
-            data_vendor = DataVendorQuandl()
-
-        elif source == 'eikon':
-            from findatapy.market.datavendorweb import DataVendorEikon
-            data_vendor = DataVendorEikon()
-
-        elif source == 'ons':
-            from findatapy.market.datavendorweb import DataVendorONS
-            data_vendor = DataVendorONS()
-
-        elif source == 'boe':
-            from findatapy.market.datavendorweb import DataVendorBOE
-            data_vendor = DataVendorBOE()
-
-        elif source == 'dukascopy':
-            from findatapy.market.datavendorweb  import DataVendorDukasCopy
-            data_vendor = DataVendorDukasCopy()
-
-        elif source == 'fxcm':
-            from findatapy.market.datavendorweb  import DataVendorFXCM
-            data_vendor = DataVendorFXCM()
-
-        elif source == 'alfred':
-            from findatapy.market.datavendorweb  import DataVendorALFRED
-            data_vendor = DataVendorALFRED()
-
-        elif source == 'yahoo':
-            from findatapy.market.datavendorweb  import DataVendorYahoo
-            data_vendor = DataVendorYahoo()
-
-        elif source in ['google', 'fred', 'oecd', 'eurostat', 'edgar-index']:
-            from findatapy.market.datavendorweb  import DataVendorPandasWeb
-            data_vendor = DataVendorPandasWeb()
-
-        elif source == 'bitcoincharts':
-            from findatapy.market.datavendorweb import DataVendorBitcoincharts
-            data_vendor = DataVendorBitcoincharts()
-        elif source == 'poloniex':
-            from findatapy.market.datavendorweb import DataVendorPoloniex
-            data_vendor = DataVendorPoloniex()
-        elif source == 'binance':
-            from findatapy.market.datavendorweb import DataVendorBinance
-            data_vendor = DataVendorBinance()
-        elif source == 'bitfinex':
-            from findatapy.market.datavendorweb import DataVendorBitfinex
-            data_vendor = DataVendorBitfinex()
-        elif source == 'gdax':
-            from findatapy.market.datavendorweb import DataVendorGdax
-            data_vendor = DataVendorGdax()
-        elif source == 'kraken':
-            from findatapy.market.datavendorweb import DataVendorKraken
-            data_vendor = DataVendorKraken()
-        elif source == 'bitmex':
-            from findatapy.market.datavendorweb import DataVendorBitmex
-            data_vendor = DataVendorBitmex()
-        elif '.csv' in source or '.h5' in source:
+        # Special case for files (csv, h5, parquet or zip)
+        if '.csv' in str(data_source) or '.h5' in str(data_source) or '.parquet' in str(data_source) or '.zip' in str(data_source) \
+                or data_engine is not None:
             from findatapy.market.datavendorweb import DataVendorFlatFile
             data_vendor = DataVendorFlatFile()
-        elif source == 'alphavantage':
-            from findatapy.market.datavendorweb import DataVendorAlphaVantage
-            data_vendor = DataVendorAlphaVantage()
-        elif source == 'huobi':
-            from findatapy.market.datavendorweb import DataVendorHuobi
-            data_vendor = DataVendorHuobi()
+        else:
+            try:
+                data_source = data_source.split("-")[0]
+            except:
+                logger.error("Was data data_source specified?")
+
+                return None
+
+            if data_source == 'bloomberg':
+                try:
+                    from findatapy.market.datavendorbbg import DataVendorBBGOpen
+                    data_vendor = DataVendorBBGOpen()
+                except:
+                    logger.warn("Bloomberg needs to be installed")
+
+            elif data_source == 'quandl':
+                from findatapy.market.datavendorweb import DataVendorQuandl
+                data_vendor = DataVendorQuandl()
+
+            elif data_source == 'eikon':
+                from findatapy.market.datavendorweb import DataVendorEikon
+                data_vendor = DataVendorEikon()
+
+            elif data_source == 'ons':
+                from findatapy.market.datavendorweb import DataVendorONS
+                data_vendor = DataVendorONS()
+
+            elif data_source == 'boe':
+                from findatapy.market.datavendorweb import DataVendorBOE
+                data_vendor = DataVendorBOE()
+
+            elif data_source == 'dukascopy':
+                from findatapy.market.datavendorweb  import DataVendorDukasCopy
+                data_vendor = DataVendorDukasCopy()
+
+            elif data_source == 'fxcm':
+                from findatapy.market.datavendorweb  import DataVendorFXCM
+                data_vendor = DataVendorFXCM()
+
+            elif data_source == 'alfred':
+                from findatapy.market.datavendorweb  import DataVendorALFRED
+                data_vendor = DataVendorALFRED()
+
+            elif data_source == 'yahoo':
+                from findatapy.market.datavendorweb  import DataVendorYahoo
+                data_vendor = DataVendorYahoo()
+
+            elif data_source in ['google', 'fred', 'oecd', 'eurostat', 'edgar-index']:
+                from findatapy.market.datavendorweb  import DataVendorPandasWeb
+                data_vendor = DataVendorPandasWeb()
+
+            elif data_source == 'bitcoincharts':
+                from findatapy.market.datavendorweb import DataVendorBitcoincharts
+                data_vendor = DataVendorBitcoincharts()
+            elif data_source == 'poloniex':
+                from findatapy.market.datavendorweb import DataVendorPoloniex
+                data_vendor = DataVendorPoloniex()
+            elif data_source == 'binance':
+                from findatapy.market.datavendorweb import DataVendorBinance
+                data_vendor = DataVendorBinance()
+            elif data_source == 'bitfinex':
+                from findatapy.market.datavendorweb import DataVendorBitfinex
+                data_vendor = DataVendorBitfinex()
+            elif data_source == 'gdax':
+                from findatapy.market.datavendorweb import DataVendorGdax
+                data_vendor = DataVendorGdax()
+            elif data_source == 'kraken':
+                from findatapy.market.datavendorweb import DataVendorKraken
+                data_vendor = DataVendorKraken()
+            elif data_source == 'bitmex':
+                from findatapy.market.datavendorweb import DataVendorBitmex
+                data_vendor = DataVendorBitmex()
+            elif data_source == 'alphavantage':
+                from findatapy.market.datavendorweb import DataVendorAlphaVantage
+                data_vendor = DataVendorAlphaVantage()
+            elif data_source == 'huobi':
+                from findatapy.market.datavendorweb import DataVendorHuobi
+                data_vendor = DataVendorHuobi()
 
         # TODO add support for other data sources (like Reuters)
 
@@ -157,8 +168,9 @@ class MarketDataGenerator(object):
         -------
         pandas.DataFrame
         """
+        logger = LoggerManager().getLogger(__name__)
 
-        # data_vendor = self.get_data_vendor(market_data_request.data_source)
+        # data_vendor = self.get_data_vendor(md_request.data_source)
 
         # check if tickers have been specified (if not load all of them for a category)
         # also handle single tickers/list tickers
@@ -169,7 +181,7 @@ class MarketDataGenerator(object):
 
         tickers = market_data_request.tickers
 
-        if tickers is None :
+        if tickers is None:
             create_tickers = True
         elif isinstance(tickers, str):
             if tickers == '': create_tickers = True
@@ -185,18 +197,18 @@ class MarketDataGenerator(object):
             data_frame_agg = self.download_intraday_tick(market_data_request)
      #       return data_frame_agg
 
-        # daily: multiple tickers per cache file - assume we make one API call to vendor library
+        # Daily: multiple tickers per cache file - assume we make one API call to vendor library
         else:
             data_frame_agg = self.download_daily(market_data_request)
 
         if('internet_load' in market_data_request.cache_algo):
-            self.logger.debug("Internet loading.. ")
+            logger.debug("Internet loading.. ")
 
             # Signal to data_vendor template to exit session
             # if data_vendor is not None and kill_session == True: data_vendor.kill_session()
 
         if(market_data_request.cache_algo == 'cache_algo'):
-            self.logger.debug("Only caching data in memory, do not return any time series."); return
+            logger.debug("Only caching data in memory, do not return any time series."); return
 
         # Only return time series if specified in the algo
         if 'return' in market_data_request.cache_algo:
@@ -210,8 +222,8 @@ class MarketDataGenerator(object):
             # TODO only do this for not daily data?
             try:
                 if data_frame_agg is not None:
-                    data_frame_agg = self.filter.filter_time_series(market_data_request, data_frame_agg, pad_columns=True)\
-                        .dropna(how = 'all')
+                    data_frame_agg = self.filter.filter_time_series(market_data_request, data_frame_agg, pad_columns=True)
+                    data_frame_agg = data_frame_agg.dropna(how='all')
 
                     # Resample data using pandas if specified in the MarketDataRequest
                     if market_data_request.resample is not None:
@@ -223,7 +235,7 @@ class MarketDataGenerator(object):
                         if 'dropna' in market_data_request.resample_how:
                             data_frame_agg = data_frame_agg.dropna(how='all')
                 else:
-                    self.logger.warn("No data returned for " + str(market_data_request.tickers))
+                    logger.warn("No data returned for " + str(market_data_request.tickers))
 
                 return data_frame_agg
             except Exception as e:
@@ -233,7 +245,7 @@ class MarketDataGenerator(object):
 
                 import traceback
 
-                self.logger.warn("No data returned for " + str(market_data_request.tickers))
+                logger.warn("No data returned for " + str(market_data_request.tickers))
 
                 return None
 
@@ -253,7 +265,8 @@ class MarketDataGenerator(object):
         if(isinstance(ticker, list)):
             ticker = ticker[0]
 
-        return self.create_cache_file_name(MarketDataRequest().create_category_key(market_data_request, ticker))
+        return self.create_cache_file_name(MarketDataRequest().create_category_key(
+            md_request=market_data_request, ticker=ticker))
 
     def download_intraday_tick(self, market_data_request):
         """Loads intraday time series from specified data provider
@@ -290,7 +303,7 @@ class MarketDataGenerator(object):
                 # data is stored on disk as float32 anyway
                 # old_finish_date = market_data_request_single.finish_date
                 #
-                # market_data_request_single.finish_date = self.refine_expiry_date(market_data_request)
+                # market_data_request_single.finish_date = self.refine_expiry_date(md_request)
                 #
                 # if market_data_request_single.finish_date >= market_data_request_single.start_date:
                 #     data_frame_single = data_vendor.load_ticker(market_data_request_single)
@@ -317,14 +330,13 @@ class MarketDataGenerator(object):
                         # else:
                         #     data_frame_agg = data_frame_single
 
-                # key = self.create_category_key(market_data_request, ticker)
+                # key = self.create_category_key(md_request, ticker)
                 # fname = self.create_cache_file_name(key)
                 # self._time_series_cache[fname] = data_frame_agg  # cache in memory (disable for intraday)
 
-
             # If you call for returning multiple tickers, be careful with memory considerations!
             if data_frame_group is not None:
-                data_frame_agg = calcuations.pandas_outer_join(data_frame_group)
+                data_frame_agg = calcuations.join(data_frame_group, how='outer')
 
             return data_frame_agg
 
@@ -345,22 +357,18 @@ class MarketDataGenerator(object):
             return self.fetch_group_time_series(market_data_request_list)
 
     def fetch_single_time_series(self, market_data_request):
-
+        logger = LoggerManager().getLogger(__name__)
         market_data_request = MarketDataRequest(md_request=market_data_request)
 
         # Only includes those tickers have not expired yet!
-        start_date = pandas.Timestamp(market_data_request.start_date).date()
+        start_date = pd.Timestamp(market_data_request.start_date).date()
 
-        import datetime
-
-        current_date = datetime.datetime.utcnow().date()
-
-        from datetime import timedelta
+        current_date = pd.Timestamp(datetime.datetime.utcnow().date())
 
         tickers = market_data_request.tickers
         vendor_tickers = market_data_request.vendor_tickers
 
-        expiry_date = market_data_request.expiry_date
+        expiry_date = pd.Timestamp(market_data_request.expiry_date)
 
         config = ConfigManager().get_instance()
 
@@ -372,25 +380,26 @@ class MarketDataGenerator(object):
                 pass
 
             if expiry_date is not None:
-                expiry_date = pandas.Timestamp(expiry_date).date()
+                expiry_date = pd.Timestamp(expiry_date)
 
-                # Use pandas Timestamp, a bit more robust with weird dates (can fail if comparing date vs datetime)
-                # if the expiry is before the start date of our download don't bother downloading this ticker
-                if expiry_date < start_date:
-                    tickers[i] = None
-
-                # Special case for futures-contracts which are intraday
-                # avoid downloading if the expiry date is very far in the past
-                # (we need this before there might be odd situations where we run on an expiry date, but still want to get
-                # data right till expiry time)
-                if market_data_request.category == 'futures-contracts' and market_data_request.freq == 'intraday' \
-                        and self.days_expired_intraday_contract_download > 0:
-
-                    if expiry_date + timedelta(days=self.days_expired_intraday_contract_download) < current_date:
+                if not(pd.isna(expiry_date)):
+                    # Use pandas Timestamp, a bit more robust with weird dates (can fail if comparing date vs datetime)
+                    # if the expiry is before the start date of our download don't bother downloading this ticker
+                    if expiry_date < start_date:
                         tickers[i] = None
 
-                if vendor_tickers is not None and tickers[i] is None:
-                    vendor_tickers[i] = None
+                    # Special case for futures-contracts which are intraday
+                    # avoid downloading if the expiry date is very far in the past
+                    # (we need this before there might be odd situations where we run on an expiry date, but still want to get
+                    # data right till expiry time)
+                    if market_data_request.category == 'futures-contracts' and market_data_request.freq == 'intraday' \
+                            and self.days_expired_intraday_contract_download > 0:
+
+                        if expiry_date + pd.Timedelta(days=self.days_expired_intraday_contract_download) < current_date:
+                            tickers[i] = None
+
+                    if vendor_tickers is not None and tickers[i] is None:
+                        vendor_tickers[i] = None
 
         market_data_request.tickers = [e for e in tickers if e != None]
 
@@ -400,7 +409,7 @@ class MarketDataGenerator(object):
         data_frame_single = None
 
         if len(market_data_request.tickers) > 0:
-            data_frame_single = self.get_data_vendor(market_data_request.data_source).load_ticker(market_data_request)
+            data_frame_single = self.get_data_vendor(market_data_request).load_ticker(market_data_request)
             #print(data_frame_single.head(n=10))
 
         if data_frame_single is not None:
@@ -408,10 +417,7 @@ class MarketDataGenerator(object):
                 data_frame_single.index.name = 'Date'
 
                 # Will fail for DataFrames which includes dates/strings (eg. futures contract names)
-                try:
-                    data_frame_single = data_frame_single.astype('float32')
-                except:
-                    self.logger.warning('Could not convert to float')
+                data_frame_single = Calculations().convert_to_numeric_dataframe(data_frame_single)
 
                 if market_data_request.freq == "second":
                     data_frame_single = data_frame_single.resample("1s")
@@ -420,6 +426,8 @@ class MarketDataGenerator(object):
 
     def fetch_group_time_series(self, market_data_request_list):
 
+        logger = LoggerManager().getLogger(__name__)
+
         data_frame_agg = None
 
         thread_no = constants.market_thread_no['other']
@@ -456,12 +464,12 @@ class MarketDataGenerator(object):
 
             if data_frame_group is not None:
                 try:
-                    data_frame_agg = self.calculations.pandas_outer_join(data_frame_group)
+                    data_frame_agg = self.calculations.join(data_frame_group, how='outer')
 
                     # Force ordering to be the same!
                     # data_frame_agg = data_frame_agg[columns]
                 except Exception as e:
-                    self.logger.warning('Possible overlap of columns? Have you specifed same ticker several times: ' + str(e))
+                    logger.warning('Possible overlap of columns? Have you specifed same ticker several times: ' + str(e))
 
         return data_frame_agg
 
@@ -478,7 +486,7 @@ class MarketDataGenerator(object):
         pandas.DataFrame
         """
 
-        key = MarketDataRequest().create_category_key(market_data_request)
+        key = MarketDataRequest().create_category_key(md_request=market_data_request)
 
         is_key_overriden = False
 
@@ -490,12 +498,14 @@ class MarketDataGenerator(object):
         # By default use other
         thread_no = constants.market_thread_no['other']
 
-        if market_data_request.data_source in constants.market_thread_no:
+        if str(market_data_request.data_source) in constants.market_thread_no:
             thread_no = constants.market_thread_no[market_data_request.data_source]
 
-        # daily data does not include ticker in the key, as multiple tickers in the same file
-        if thread_no == 1:
-            # data_frame_agg = data_vendor.load_ticker(market_data_request)
+        # Daily data does not include ticker in the key, as multiple tickers in the same file
+        if thread_no == 1 or '.csv' in str(market_data_request.data_source) or \
+            '.h5' in str(market_data_request.data_source) or '.parquet' in str(market_data_request.data_source) \
+                or '.zip' in str(market_data_request.data_source) or market_data_request.data_engine is not None:
+            # data_frame_agg = data_vendor.load_ticker(md_request)
             data_frame_agg = self.fetch_single_time_series(market_data_request)
         else:
             market_data_request_list = []
@@ -524,7 +534,7 @@ class MarketDataGenerator(object):
                 for md in market_data_request_list:
                     data_frame_list.append(self.fetch_single_time_series(md))
 
-                data_frame_agg = self.calculations.pandas_outer_join(data_frame_list)
+                data_frame_agg = self.calculations.join(data_frame_list, how='outer')
             else:
                 data_frame_agg = self.fetch_group_time_series(market_data_request_list)
 
