@@ -64,11 +64,11 @@ class MarketDataRequest(object):
         else:
             ticker = self.tickers[0]
 
-        self.__category_key = self.create_category_key(self, ticker=ticker)
+        self.__category_key = self.create_category_key(md_request=self, ticker=ticker)
 
         return SpeedCache().generate_key(self, ['logger', '_MarketDataRequest__abstract_curve',
                                                 '_MarketDataRequest__cache_algo',
-                                                '_MarketDataRequest__overrides'])
+                                                '_MarketDataRequest__overrides']) + "_df"
 
     def __init__(self, data_source=None,
                  start_date='year', finish_date=datetime.datetime.utcnow(),
@@ -76,15 +76,27 @@ class MarketDataRequest(object):
                  gran_freq=None, cut="NYC",
                  fields=['close'], cache_algo="internet_load_return",
                  vendor_tickers=None, vendor_fields=None,
-                 environment="backtest", trade_side='trade', expiry_date=None, resample=None, resample_how='last',
+                 environment=data_constants.default_data_environment, trade_side='trade', expiry_date=None, resample=None, resample_how='last',
+
+                 split_request_chunks=0,
+                 list_threads=1,
+
+                 fx_vol_part=data_constants.fx_vol_part, fx_vol_tenor=data_constants.fx_vol_tenor,
+                 fx_forwards_tenor=data_constants.fx_forwards_tenor,
+                 base_depos_currencies=data_constants.base_depos_currencies,
+                 base_depos_tenor=data_constants.base_depos_tenor,
+
+                 data_engine = data_constants.default_data_engine,
+
                  md_request=None, abstract_curve=None, quandl_api_key=data_constants.quandl_api_key,
                  fred_api_key=data_constants.fred_api_key, alpha_vantage_api_key=data_constants.alpha_vantage_api_key,
                  eikon_api_key=data_constants.eikon_api_key,
                  push_to_cache=True,
-                 overrides={}
+                 overrides={},
+                 freeform_md_request={}
                  ):
 
-        # can deep copy MarketDataRequest (use a lock, so can be used with threading when downloading time series)
+        # Can deep copy MarketDataRequest (use a lock, so can be used with threading when downloading time series)
         if md_request is not None:
             import threading
             lock = threading.Lock()
@@ -118,6 +130,18 @@ class MarketDataRequest(object):
                 self.expiry_date = copy.deepcopy(md_request.expiry_date)
                 self.resample = copy.deepcopy(md_request.resample)
                 self.resample_how = copy.deepcopy(md_request.resample_how)
+
+                self.split_request_chunks = copy.deepcopy(md_request.split_request_chunks)
+                self.list_threads = copy.deepcopy(md_request.list_threads)
+
+                self.fx_vol_part = copy.deepcopy(md_request.fx_vol_part)
+                self.fx_vol_tenor = copy.deepcopy(md_request.fx_vol_tenor)
+                self.fx_forwards_tenor = copy.deepcopy(md_request.fx_forwards_tenor)
+                self.base_depos_currencies = copy.deepcopy(md_request.base_depos_currencies)
+                self.base_depos_tenor = copy.deepcopy(md_request.base_depos_tenor)
+                
+                self.data_engine = copy.deepcopy(md_request.data_engine)
+
                 self.abstract_curve = copy.deepcopy(md_request.abstract_curve)
                 self.quandl_api_key = copy.deepcopy(md_request.quandl_api_key)
                 self.fred_api_key = copy.deepcopy(md_request.fred_api_key)
@@ -125,8 +149,9 @@ class MarketDataRequest(object):
                 self.eikon_api_key = copy.deepcopy(md_request.eikon_api_key)
                 self.overrides = copy.deepcopy(md_request.overrides)
                 self.push_to_cache = copy.deepcopy(md_request.push_to_cache)
+                self.freeform_md_request = copy.deepcopy(md_request.freeform_md_request)
 
-                self.tickers = copy.deepcopy(md_request.tickers)  # need this after category in case have wildcard
+                self.tickers = copy.deepcopy(md_request.tickers)  # Need this after category in case have wildcard
         else:
             self.freq_mult = freq_mult
 
@@ -151,6 +176,18 @@ class MarketDataRequest(object):
             self.expiry_date = expiry_date
             self.resample = resample
             self.resample_how = resample_how
+            
+            self.split_request_chunks = split_request_chunks
+            self.list_threads = list_threads
+
+            self.fx_vol_part = fx_vol_part
+            self.fx_vol_tenor = fx_vol_tenor
+            self.fx_forwards_tenor = fx_forwards_tenor
+            self.base_depos_currencies = base_depos_currencies
+            self.base_depos_tenor = base_depos_tenor
+            
+            self.data_engine = data_engine
+
             self.abstract_curve = abstract_curve
             
             self.quandl_api_key = quandl_api_key
@@ -161,15 +198,25 @@ class MarketDataRequest(object):
             self.overrides = overrides
             self.push_to_cache = push_to_cache
 
+            self.freeform_md_request = freeform_md_request
+
             self.tickers = tickers
 
-    def create_category_key(self, market_data_request, ticker=None):
+            if self.tickers is None:
+                self.tickers = vendor_tickers
+
+            self.old_tickers = self.tickers
+
+    def __str__(self):
+        return "MarketDataRequest summary - " + self.generate_key()
+
+    def create_category_key(self, md_request=None, ticker=None):
         """Returns a category key for the associated MarketDataRequest, which can be used to create filenames (or
         as part of a storage key in a cache)
 
         Parameters
         ----------
-        market_data_request : MarketDataRequest
+        md_request : MarketDataRequest
             contains various properties describing time series to fetched, including ticker, start & finish date etc.
 
         Returns
@@ -180,13 +227,19 @@ class MarketDataRequest(object):
         category = 'default-cat'
         cut = 'default-cut'
 
-        if market_data_request.category is not None: category = market_data_request.category
+        if md_request is None:
+            md_request = self
 
-        environment = market_data_request.environment
-        source = market_data_request.data_source
-        freq = market_data_request.freq
+        if md_request.category is not None: category = md_request.category
 
-        if market_data_request.cut is not None: cut = market_data_request.cut
+        environment = md_request.environment
+        source = md_request.data_source
+        freq = md_request.freq
+
+        if ticker is None and (md_request.freq == 'intraday' or md_request.freq == 'tick'):
+            ticker = md_request.tickers[0]
+
+        if md_request.cut is not None: cut = md_request.cut
 
         if (ticker is not None):
             key = str(environment) + "." + str(category) + '.' + str(source) + '.' + str(freq) + '.' + str(cut) \
@@ -259,7 +312,16 @@ class MarketDataRequest(object):
             self.__tickers = new_tickers
         else:
             self.__tickers = tickers
+    
+    
+    @property
+    def old_tickers(self):
+        return self.__old_tickers
 
+    @old_tickers.setter
+    def old_tickers(self, old_tickers):
+        self.__old_tickers = old_tickers
+        
     @property
     def fields(self):
         return self.__fields
@@ -393,25 +455,41 @@ class MarketDataRequest(object):
     @resample_how.setter
     def resample_how(self, resample_how):
         self.__resample_how = resample_how
+        
+    @property
+    def split_request_chunks(self):
+        return self.__split_request_chunks
+
+    @split_request_chunks.setter
+    def split_request_chunks(self, split_request_chunks):
+        self.__split_request_chunks = split_request_chunks
+        
+    @property
+    def list_threads(self):
+        return self.__list_threads
+
+    @list_threads.setter
+    def list_threads(self, list_threads):
+        self.__list_threads = list_threads
 
     def date_parser(self, date):
         if isinstance(date, str):
 
             date1 = datetime.datetime.utcnow()
 
-            if date is 'midnight':
+            if date == 'midnight':
                 date1 = datetime.datetime(date1.year, date1.month, date1.day, 0, 0, 0)
-            elif date is 'decade':
+            elif date == 'decade':
                 date1 = date1 - timedelta(days=365 * 10)
-            elif date is 'year':
+            elif date == 'year':
                 date1 = date1 - timedelta(days=365)
-            elif date is 'month':
+            elif date == 'month':
                 date1 = date1 - timedelta(days=30)
-            elif date is 'week':
+            elif date == 'week':
                 date1 = date1 - timedelta(days=7)
-            elif date is 'day':
+            elif date == 'day':
                 date1 = date1 - timedelta(days=1)
-            elif date is 'hour':
+            elif date == 'hour':
                 date1 = date1 - timedelta(hours=1)
             else:
                 # format expected 'Jun 1 2005 01:33', '%b %d %Y %H:%M'
@@ -428,6 +506,13 @@ class MarketDataRequest(object):
                     # self.logger.warning("Attempted to parse date")
                     i = 0
 
+                # format expected '1 June 2005 01:33', '%d %B %Y %H:%M'
+                try:
+                    date1 = datetime.datetime.strptime(date, '%d %B %Y %H:%M')
+                except:
+                    # self.logger.warning("Attempted to parse date")
+                    i = 0
+
                 try:
                     date1 = datetime.datetime.strptime(date, '%b %d %Y')
                 except:
@@ -439,6 +524,18 @@ class MarketDataRequest(object):
                 except:
                     # self.logger.warning("Attempted to parse date")
                     i = 0
+
+                try:
+                    date1 = datetime.datetime.strptime(date, '%B %d %Y')
+                except:
+                    # self.logger.warning("Attempted to parse date")
+                    i = 0
+
+                try:
+                    date1 = datetime.datetime.strptime(date, '%d %B %Y')
+                except:
+                    # self.logger.warning("Attempted to parse date")
+                    i = 0
         else:
             import pandas
 
@@ -469,7 +566,7 @@ class MarketDataRequest(object):
     def environment(self, environment):
         environment = environment.lower()
 
-        valid_environment = ['prod', 'backtest']
+        valid_environment = data_constants.possible_data_environment
 
         if not environment in valid_environment:
             LoggerManager().getLogger(__name__).warning(environment + " is not a defined environment.")
@@ -498,7 +595,57 @@ class MarketDataRequest(object):
     @expiry_date.setter
     def expiry_date(self, expiry_date):
         self.__expiry_date = self.date_parser(expiry_date)
+        
+    ###### For FX vol, FX forwards and base depos ######
+    @property
+    def fx_vol_part(self):
+        return self.__fx_vol_part
 
+    @fx_vol_part.setter
+    def fx_vol_part(self, fx_vol_part):
+        self.__fx_vol_part = fx_vol_part
+        
+    @property
+    def fx_vol_tenor(self):
+        return self.__fx_vol_tenor
+
+    @fx_vol_tenor.setter
+    def fx_vol_tenor(self, fx_vol_tenor):
+        self.__fx_vol_tenor = fx_vol_tenor
+        
+    @property
+    def fx_forwards_tenor(self):
+        return self.__fx_forwards_tenor
+
+    @fx_forwards_tenor.setter
+    def fx_forwards_tenor(self, fx_forwards_tenor):
+        self.__fx_forwards_tenor = fx_forwards_tenor
+
+    @property
+    def base_depos_currencies(self):
+        return self.__base_depos_currencies
+
+    @base_depos_currencies.setter
+    def base_depos_currencies(self, base_depos_currencies):
+        self.__base_depos_currencies = base_depos_currencies
+        
+    @property
+    def base_depos_tenor(self):
+        return self.__base_depos_tenor
+
+    @base_depos_tenor.setter
+    def base_depos_tenor(self, base_depos_tenor):
+        self.__base_depos_tenor = base_depos_tenor
+        
+    @property
+    def data_engine(self):
+        return self.__data_engine
+
+    @data_engine.setter
+    def data_engine(self, data_engine):
+        self.__data_engine = data_engine
+
+    ######
     @property
     def abstract_curve(self):
         return self.__abstract_curve
@@ -559,6 +706,14 @@ class MarketDataRequest(object):
     @push_to_cache.setter
     def push_to_cache(self, push_to_cache):
         self.__push_to_cache = push_to_cache
+        
+    @property
+    def freeform_md_request(self):
+        return self.__freeform_md_request
+
+    @freeform_md_request.setter
+    def freeform_md_request(self, freeform_md_request):
+        self.__freeform_md_request = freeform_md_request
 
     def _flatten_list(self, list_of_lists):
         """Flattens list, particularly useful for combining baskets
