@@ -1,7 +1,7 @@
-__author__ = 'saeedamen' # Saeed Amen
+__author__ = 'saeedamen'  # Saeed Amen
 
 #
-# Copyright 2016 Cuemacro
+# Copyright 2016-2020 Cuemacro
 #
 # Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the
 # License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
@@ -12,29 +12,38 @@ __author__ = 'saeedamen' # Saeed Amen
 # See the License for the specific language governing permissions and limitations under the License.
 #
 
+import numpy as np
+import pandas as pd
+import pytz
+
+import datetime
+from datetime import timedelta
+
+from findatapy.timeseries.calendar import Calendar
+
+
+from findatapy.util.dataconstants import DataConstants
 from findatapy.util.loggermanager import LoggerManager
 
-from pandas.tseries.offsets import CustomBusinessDay
-
-import numpy as np
-import pandas
-import pytz
+constants = DataConstants()
 
 class Filter(object):
     """Functions for filtering time series by dates and columns.
 
     This class is used extensively in both findatapy and finmarketpy.
 
+    Market holidays are collected from web sources such as https://www.timeanddate.com/holidays/ and also individual
+    exchange websites, and is manually updated from time to time to take into account newly instituted holidays, and stored
+    in conf/holidays_table.parquet - if you need to add your own holidays.
+
     """
 
-    _time_series_cache = {} # shared across all instances of object!
+    _time_series_cache = {}  # shared across all instances of object!
 
     def __init__(self):
-        # self.config = ConfigManager()
-        self.logger = LoggerManager().getLogger(__name__)
-        return
+        self._calendar = Calendar()
 
-    def filter_time_series(self, market_data_request, data_frame, pad_columns = False):
+    def filter_time_series(self, market_data_request, data_frame, pad_columns=False):
         """Filters a time series given a set of criteria (like start/finish date and tickers)
 
         Parameters
@@ -55,7 +64,7 @@ class Filter(object):
 
         data_frame = self.filter_time_series_by_date(start_date, finish_date, data_frame)
 
-        # filter by ticker.field combinations requested
+        # Filter by ticker.field combinations requested
         columns = self.create_tickers_fields_list(market_data_request)
 
         if (pad_columns):
@@ -65,71 +74,7 @@ class Filter(object):
 
         return data_frame
 
-    def create_calendar_bus_days(self, start_date, end_date, cal = 'FX'):
-        """Creates a calendar of business days
-
-        Parameters
-        ----------
-        start_date : DateTime
-            start date of calendar
-        end_date : DataFrame
-            finish date of calendar
-        cal : str
-            business calendar to use
-
-        Returns
-        -------
-        list
-        """
-        hols = self.get_holidays(start_date, end_date, cal)
-        index = pandas.bdate_range(start=start_date, end=end_date, freq='D')
-
-        return [x for x in index if x not in hols]
-
-    def get_holidays(self, start_date, end_date, cal = 'FX', holidays_list = []):
-        """Gets the holidays for a given calendar
-
-        Parameters
-        ----------
-        start_date : DateTime
-            start date of calendar
-        end_date : DataFrame
-            finish date of calendar
-        cal : str
-            business calendar to use
-
-        Returns
-        -------
-        list
-        """
-
-        # TODO use Pandas CustomBusinessDays to get more calendars
-
-        if cal == 'FX':
-            # filter for Christmas & New Year's Day
-            for i in range(1970, 2020):
-                holidays_list.append(str(i) + "-12-25")
-                holidays_list.append(str(i) + "-01-01")
-
-        if cal == 'WEEKDAY':
-            bday = CustomBusinessDay(weekmask='Sat Sun')
-
-            holidays_list = pandas.date_range(start_date, end_date, freq=bday)
-
-        # holidays_list = pandas.to_datetime(holidays_list).order()
-        holidays_list = pandas.to_datetime(holidays_list).sort_values()
-
-        # floor start date
-        start = np.datetime64(start_date) - np.timedelta64(1, 'D')
-
-        # ceiling end date
-        end = np.datetime64(end_date) + np.timedelta64(1, 'D')
-
-        holidays_list = [x for x in holidays_list if x >= start and x <= end]
-
-        return pandas.to_datetime(holidays_list).tz_localize('UTC')
-
-    def filter_time_series_by_holidays(self, data_frame, cal = 'FX', holidays_list = []):
+    def filter_time_series_by_holidays(self, data_frame, cal='FX', holidays_list=[]):
         """Removes holidays from a given time series
 
         Parameters
@@ -144,17 +89,17 @@ class Filter(object):
         DataFrame
         """
 
-        # optimal case for weekdays: remove Saturday and Sunday
-        if (cal == 'WEEKDAY'):
+        # Optimal case for weekdays: remove Saturday and Sunday
+        if (cal == 'WEEKDAY' or cal == 'WKY'):
             return data_frame[data_frame.index.dayofweek <= 4]
 
-        # select only those holidays in the sample
-        holidays_start = self.get_holidays(data_frame.index[0], data_frame.index[-1], cal, holidays_list = holidays_list)
+        # Select only those holidays in the sample
+        holidays_start = self._calendar.get_holidays(data_frame.index[0], data_frame.index[-1], cal, holidays_list=holidays_list)
 
-        if(holidays_start.size == 0):
+        if (holidays_start.size == 0):
             return data_frame
 
-        holidays_end = holidays_start + np.timedelta64(1,'D')
+        holidays_end = holidays_start + np.timedelta64(1, 'D')
 
         # floored_dates = data_frame.index.normalize()
         #
@@ -189,8 +134,6 @@ class Filter(object):
         data_frame_filtered = []
 
         for i in range(0, len(holidays_start)):
-
-
             data_frame_temp = data_frame_left[data_frame_left.index < holidays_start[i]]
             data_frame_left = data_frame_left[data_frame_left.index >= holidays_end[i]]
 
@@ -198,7 +141,7 @@ class Filter(object):
 
         data_frame_filtered.append(data_frame_left)
 
-        return pandas.concat(data_frame_filtered)
+        return pd.concat(data_frame_filtered)
 
     def filter_time_series_by_date(self, start_date, finish_date, data_frame):
         """Filter time series by start/finish dates
@@ -216,9 +159,10 @@ class Filter(object):
         -------
         DataFrame
         """
-        offset = 0 # inclusive
+        offset = 0  # inclusive
 
-        return self.filter_time_series_by_date_offset(start_date, finish_date, data_frame, offset, exclude_start_end = False)
+        return self.filter_time_series_by_date_offset(start_date, finish_date, data_frame, offset,
+                                                      exclude_start_end=False)
 
     def filter_time_series_by_days(self, days, data_frame):
         """Filter time series by start/finish dates
@@ -236,10 +180,11 @@ class Filter(object):
         -------
         DataFrame
         """
-        offset = 0 # inclusive
+        offset = 0  # inclusive
 
         finish_date = datetime.datetime.utcnow()
         start_date = finish_date - timedelta(days=days)
+
         return self.filter_time_series_by_date_offset(start_date, finish_date, data_frame, offset)
 
     def filter_time_series_by_date_exc(self, start_date, finish_date, data_frame):
@@ -258,9 +203,10 @@ class Filter(object):
         -------
         DataFrame
         """
-        offset = 1 # exclusive of start finish date
+        offset = 1  # exclusive of start finish date
 
-        return self.filter_time_series_by_date_offset(start_date, finish_date, data_frame, offset, exclude_start_end = True)
+        return self.filter_time_series_by_date_offset(start_date, finish_date, data_frame, offset,
+                                                      exclude_start_end=True)
 
         # try:
         #     # filter by dates for intraday data
@@ -281,7 +227,7 @@ class Filter(object):
         #
         # return data_frame
 
-    def filter_time_series_by_date_offset(self, start_date, finish_date, data_frame, offset, exclude_start_end = False):
+    def filter_time_series_by_date_offset(self, start_date, finish_date, data_frame, offset, exclude_start_end=False):
         """Filter time series by start/finish dates (and an offset)
 
         Parameters
@@ -311,11 +257,17 @@ class Filter(object):
                     finish_date = finish_date.replace(tzinfo=data_frame.index.tz)
             else:
                 # Otherwise remove timezone from start_date/finish_date
-                if not (isinstance(start_date, str)):
-                    start_date = start_date.replace(tzinfo=None)
+                if not(isinstance(start_date, str)):
+                    try:
+                        start_date = start_date.replace(tzinfo=None)
+                    except:
+                        pass
 
                 if not (isinstance(finish_date, str)):
-                    finish_date = finish_date.replace(tzinfo=None)
+                    try:
+                        finish_date = finish_date.replace(tzinfo=None)
+                    except:
+                        pass
 
         if 'int' in str(data_frame.index.dtype):
             return data_frame
@@ -349,13 +301,13 @@ class Filter(object):
 
             # if we have dates stored as opposed to TimeStamps (ie. daily data), we use a simple (slower) method
             # for filtering daily data
-            if(start_date is not None):
+            if (start_date is not None):
                 if exclude_start_end:
                     data_frame = data_frame.loc[start_date < data_frame.index]
                 else:
                     data_frame = data_frame.loc[start_date <= data_frame.index]
 
-            if(finish_date is not None):
+            if (finish_date is not None):
                 if exclude_start_end:
                     data_frame = data_frame.loc[data_frame.index < finish_date]
                 else:
@@ -413,8 +365,19 @@ class Filter(object):
         # Much faster, start and finish dates are inclusive
         return data_frame[(data_frame.index >= start_date) & (data_frame.index <= finish_date)]
 
+    def filter_time_series_by_time_of_day_timezone(self, hour, minute, data_frame, timezone_of_snap='UTC'):
 
-    def filter_time_series_by_time_of_day(self, hour, minute, data_frame, in_tz = None, out_tz = None):
+        old_tz = data_frame.index.tz
+        data_frame = data_frame.tz_convert(pytz.timezone(timezone_of_snap))
+
+        data_frame = data_frame[data_frame.index.minute == minute]
+        data_frame = data_frame[data_frame.index.hour == hour]
+
+        data_frame = data_frame.tz_convert(old_tz)
+
+        return data_frame
+
+    def filter_time_series_by_time_of_day(self, hour, minute, data_frame, in_tz=None, out_tz=None):
         """Filter time series by time of day
 
         Parameters
@@ -435,20 +398,23 @@ class Filter(object):
         DataFrame
         """
         if out_tz is not None:
-            if in_tz is not None:
-                data_frame = data_frame.tz_localize(pytz.timezone(in_tz))
+            try:
+                if in_tz is not None:
+                    data_frame = data_frame.tz_localize(pytz.timezone(in_tz))
+            except:
+                data_frame = data_frame.tz_convert(pytz.timezone(in_tz))
 
             data_frame = data_frame.tz_convert(pytz.timezone(out_tz))
 
             # change internal representation of time
-            data_frame.index = pandas.DatetimeIndex(data_frame.index.values)
+            data_frame.index = pd.DatetimeIndex(data_frame.index.values)
 
         data_frame = data_frame[data_frame.index.minute == minute]
         data_frame = data_frame[data_frame.index.hour == hour]
 
         return data_frame
 
-    def filter_time_series_by_minute_of_hour(self, minute, data_frame, in_tz = None, out_tz = None):
+    def filter_time_series_by_minute_of_hour(self, minute, data_frame, in_tz=None, out_tz=None):
         """Filter time series by minute of hour
 
         Parameters
@@ -473,7 +439,7 @@ class Filter(object):
             data_frame = data_frame.tz_convert(pytz.timezone(out_tz))
 
             # change internal representation of time
-            data_frame.index = pandas.DatetimeIndex(data_frame.index.values)
+            data_frame.index = pd.DatetimeIndex(data_frame.index.values)
 
         data_frame = data_frame[data_frame.index.minute == minute]
 
@@ -534,7 +500,7 @@ class Filter(object):
         -------
         DataFrame
         """
-        old_columns = data_frame.columns
+        old_columns = data_frame.columns.tolist()
 
         common_columns = [val for val in columns if val in old_columns]
         uncommon_columns = [val for val in columns if val not in old_columns]
@@ -543,16 +509,32 @@ class Filter(object):
         data_frame = data_frame[common_columns]
 
         if len(uncommon_columns) > 0:
-            self.logger.info("Padding missing columns...") # " + str(uncommon_columns))
+            logger = LoggerManager().getLogger(__name__)
 
-            new_data_frame = pandas.DataFrame(index=data_frame.index, columns=uncommon_columns)
+            logger.info("Padding missing columns...")  # " + str(uncommon_columns))
 
-            data_frame = pandas.concat([data_frame, new_data_frame], axis=1)
+            new_data_frame = pd.DataFrame(index=data_frame.index, columns=uncommon_columns)
+
+            data_frame = pd.concat([data_frame, new_data_frame], axis=1)
+
+            # Force new columns to float NaNs (not objects which causes problems with newer pandas versions)
+            # or to NaT if they are date columns
+            for u in uncommon_columns:
+                is_date = False
+
+                for c in constants.always_date_columns:
+                    if c in u:
+                        is_date = True
+
+                if is_date:
+                    data_frame[u] = np.datetime64('NaT')
+                else:
+                    data_frame[u] = np.nan
 
             # SLOW method below
             # for x in uncommon_columns: data_frame.loc[:,x] = np.nan
 
-        # get columns in same order again
+        # Get columns in same order again
         data_frame = data_frame[columns]
 
         return data_frame
@@ -572,7 +554,7 @@ class Filter(object):
         DataFrame
         """
 
-        if not(isinstance(keyword, list)):
+        if not (isinstance(keyword, list)):
             keyword = [keyword]
 
         columns = []
@@ -580,11 +562,11 @@ class Filter(object):
         for k in keyword:
             columns.append([elem for elem in data_frame.columns if k not in elem])
 
-        columns = self.flatten_list_of_lists(columns)
+        columns = self._calendar.flatten_list_of_lists(columns)
 
         return self.filter_time_series_by_columns(columns, data_frame)
 
-    def filter_time_series_by_included_keyword(self, keyword, data_frame):
+    def filter_time_series_by_included_keyword(self, keyword, data_frame, ignore_case=False):
         """Filter time series to include columns which contain keyword
 
         Parameters
@@ -599,43 +581,22 @@ class Filter(object):
         DataFrame
         """
 
-        if not(isinstance(keyword, list)):
+        if not (isinstance(keyword, list)):
             keyword = [keyword]
 
         columns = []
 
-        for k in keyword:
-            columns.append([elem for elem in data_frame.columns if k in elem])
+        if ignore_case:
+            for k in keyword:
+                columns.append([elem for elem in data_frame.columns if k.lower() in elem.lower()])
+        else:
+            for k in keyword:
+                columns.append([elem for elem in data_frame.columns if k in elem])
 
-        columns = self.flatten_list_of_lists(columns)
+        columns = self._calendar.flatten_list_of_lists(columns)
 
         return self.filter_time_series_by_columns(columns, data_frame)
 
-    def flatten_list_of_lists(self, list_of_lists):
-        """Flattens lists of obj, into a single list of strings (rather than characters, which is default behavior).
-
-        Parameters
-        ----------
-        list_of_lists : obj (list)
-            List to be flattened
-
-        Returns
-        -------
-        str (list)
-        """
-
-        if isinstance(list_of_lists, list):
-            rt = []
-            for i in list_of_lists:
-                if isinstance(i, list):
-                    rt.extend(self.flatten_list_of_lists(i))
-                else:
-                    rt.append(i)
-
-            return rt
-
-        return list_of_lists
-
     def filter_time_series_by_minute_freq(self, freq, data_frame):
         """Filter time series where minutes correspond to certain minute filter
 
@@ -680,10 +641,10 @@ class Filter(object):
         return tickers_fields_list
 
     def resample_time_series(self, data_frame, freq):
-        return data_frame.asfreq(freq, method = 'pad')
+        return data_frame.asfreq(freq, method='pad')
 
     def resample_time_series_frequency(self, data_frame, data_resample_freq,
-                                       data_resample_type = 'mean', fill_empties = False):
+                                       data_resample_type='mean', fill_empties=False):
         # Should we take the mean, first, last in our resample
         if data_resample_type == 'mean':
             data_frame_r = data_frame.resample(data_resample_freq).mean()
@@ -729,7 +690,7 @@ class Filter(object):
         # Monday = 0, ..., Sunday = 6
         data_frame = data_frame[~((data_frame.index.dayofweek == 4) & (data_frame.index.hour > 22))]
         data_frame = data_frame[~((data_frame.index.dayofweek == 5))]
-        data_frame = data_frame[~((data_frame.index.dayofweek == 6)& (data_frame.index.hour < 19))]
+        data_frame = data_frame[~((data_frame.index.dayofweek == 6) & (data_frame.index.hour < 19))]
         data_frame = data_frame[~((data_frame.index.day == 1) & (data_frame.index.month == 1))]
 
         return data_frame
@@ -737,7 +698,6 @@ class Filter(object):
     def remove_duplicate_indices(self, df):
         return df[~df.index.duplicated(keep='first')]
 
-
     def mask_time_series_by_time(self, df, time_list, time_zone):
         """ Mask a time series by time of day and time zone specified
         e.g. given a time series minutes data
@@ -761,7 +721,7 @@ class Filter(object):
 
         # Change the time zone from 'UTC' to a given one
         df.index = df.index.tz_convert(time_zone)
-        df_mask = pandas.DataFrame(0,index=df.index,columns=['mask'])
+        df_mask = pd.DataFrame(0, index=df.index, columns=['mask'])
 
         # Mask data with each given tuple
         for i in range(0, len(time_list)):
@@ -772,174 +732,22 @@ class Filter(object):
 
             # E.g. if tuple is ('01:08', '03:02'),
             # take hours in target - take values in [01:00,04:00]
-            narray = np.where(df.index.hour.isin(range(start_hour,end_hour + 1)), 1, 0)
-            df_mask_temp = pandas.DataFrame(index=df.index, columns=df_mask.columns.tolist(), data=narray)
+            narray = np.where(df.index.hour.isin(range(start_hour, end_hour + 1)), 1, 0)
+            df_mask_temp = pd.DataFrame(index=df.index, columns=df_mask.columns.tolist(), data=narray)
 
             # Remove minutes not in target - remove values in [01:00,01:07], [03:03,03:59]
             narray = np.where(((df.index.hour == start_hour) & (df.index.minute < start_minute)), 0, 1)
-            df_mask_temp = df_mask_temp * pandas.DataFrame(index=df.index, columns=df_mask.columns.tolist(), data=narray)
+            df_mask_temp = df_mask_temp * pd.DataFrame(index=df.index, columns=df_mask.columns.tolist(),
+                                                           data=narray)
             narray = np.where((df.index.hour == end_hour) & (df.index.minute > end_minute), 0, 1)
-            df_mask_temp = df_mask_temp * pandas.DataFrame(index=df.index, columns=df_mask.columns.tolist(), data=narray)
+            df_mask_temp = df_mask_temp * pd.DataFrame(index=df.index, columns=df_mask.columns.tolist(),
+                                                           data=narray)
 
             # Collect all the periods we want to keep the data
             df_mask = df_mask + df_mask_temp
 
         narray = np.where(df_mask == 1, df, 0)
-        df = pandas.DataFrame(index=df.index,columns=df.columns.tolist(),data=narray)
+        df = pd.DataFrame(index=df.index, columns=df.columns.tolist(), data=narray)
         df.index = df.index.tz_convert('UTC')  # change the time zone to 'UTC'
 
         return df
-
-#######################################################################################################################
-
-import datetime
-from datetime import timedelta
-
-import numpy
-import pandas
-import pandas.tseries.offsets
-
-from findatapy.timeseries.timezone import Timezone
-
-from pandas.tseries.offsets import BDay
-from pandas.tseries.offsets import CustomBusinessDay
-
-from findatapy.timeseries.filter import Filter
-
-class Calendar(object):
-    """Provides calendar based functions for working out options expiries. Note, that in practice, we would often take
-    into account market holidays.
-
-    """
-
-    def get_business_days_tenor(self, tenor):
-        if tenor == '1W':
-            return 5
-        elif tenor == 'ON':
-            return 1
-        elif tenor == '1M':
-            return 20
-        elif tenor == '3M':
-            return 60
-        elif tenor == '6M':
-            return 120
-        elif tenor == '1Y':
-            return 252
-
-    def get_dates_from_tenors(self, start, end, calendar, tenor):
-        freq = str(self.get_business_days_tenor(tenor)) + "B"
-        return pandas.DataFrame(index=pandas.bdate_range(start, end, freq=freq))
-
-    def get_expiries_from_dates(self, date_time_index, calendar, tenor):
-        freq = self.get_business_days_tenor(tenor)
-
-        return pandas.DatetimeIndex(date_time_index + BDay(freq))
-
-    def align_to_NY_cut_in_UTC(self, date_time):
-
-        tstz = Timezone()
-        date_time = tstz.localise_index_as_new_york_time(date_time)
-        date_time.index = date_time.index + timedelta(hours=10)
-
-        return tstz.convert_index_aware_to_UTC_time(date_time)
-
-    def floor_date(self, data_frame):
-        data_frame.index = data_frame.index.normalize()
-
-        return data_frame
-
-    def create_bus_day(self, start, end):
-        return pandas.date_range(start, end, freq='B')
-
-    def get_bus_day_of_month(self, date, cal = 'FX'):
-        """ get_bus_day_of_month(date = list of dates, cal = calendar name)
-
-            returns the business day of the month (ie. 3rd Jan, on a Monday,
-            would be the 1st business day of the month
-        """
-        filter = Filter()
-
-        try:
-            date = date.normalize() # strip times off the dates - for business dates just want dates!
-        except: pass
-
-        start = pandas.to_datetime(datetime.datetime(date.year[0], date.month[0], 1))
-        end = datetime.datetime.today()#pandas.to_datetime(datetime.datetime(date.year[-1], date.month[-1], date.day[-1]))
-
-        holidays = filter.get_holidays(start, end, cal)
-
-        bday = CustomBusinessDay(holidays=holidays, weekmask='Mon Tue Wed Thu Fri')
-
-        bus_dates = pandas.date_range(start, end, freq=bday)
-
-        month = bus_dates.month
-
-        work_day_index = numpy.zeros(len(bus_dates))
-        work_day_index[0] = 1
-
-        for i in range(1, len(bus_dates)):
-            if month[i] == month[i-1]:
-                work_day_index[i] = work_day_index[i-1] + 1
-            else:
-                work_day_index[i] = 1
-
-        bus_day_of_month = work_day_index[bus_dates.searchsorted(date)]
-
-        return bus_day_of_month
-
-    def set_market_holidays(self, holiday_df):
-        self.holiday_df = holiday_df
-
-# functions to test class
-if __name__ == '__main__':
-
-    logger = LoggerManager.getLogger(__name__)
-
-    tsf = Filter()
-
-    if True:
-        import pandas as pd
-        dates = pd.date_range('01 Jan 2020', '10 Jan 2020', freq='D')
-        data_frame = pd.DataFrame(index=dates)
-
-        start_date = '03 Jan 2020'; finish_date = '06 Jan 2020'
-        print(data_frame.loc[start_date:finish_date])
-
-        # Much faster!
-        print(data_frame[(data_frame.index >= start_date) & (data_frame.index <= finish_date)])
-
-    if False:
-        start = pandas.to_datetime('2000-01-01')
-        end = pandas.to_datetime('2020-01-01')
-
-        logger.info('Get FX holidays')
-        hols = tsf.get_holidays(start, end, cal='FX')
-        print(hols)
-
-        logger.info('Get business days, excluding holidays')
-        bus_days = tsf.create_calendar_bus_days(start, end, cal='FX')
-        print(bus_days)
-
-    if False:
-        logger.info('Remove out of hours')
-
-        rng = pandas.date_range('01 Jan 2014', '05 Jan 2014', freq='1min')
-        intraday_vals = pandas.DataFrame(data=pandas.np.random.randn(len(rng)), index=rng)
-
-        intraday_vals = tsf.resample_time_series(intraday_vals, '60min')
-        intraday_vals = tsf.remove_out_FX_out_of_hours(intraday_vals)
-
-        print(intraday_vals)
-
-    if True:
-        logger.info('Remove holiday days')
-
-        rng = pandas.date_range('01 Jan 2007', '05 Jan 2014', freq='1min')
-        intraday_vals = pandas.DataFrame(data=pandas.np.random.randn(len(rng)), index=rng)
-
-        import cProfile
-
-        cProfile.run("intraday_vals = tsf.filter_time_series_by_holidays(intraday_vals, 'FX')")
-
-        print(intraday_vals)
-
